from pathlib import Path

from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils import translation
from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import fs_cleanup
from mayan.apps.testing.tests.base import BaseTestCase
from mayan.apps.views.settings import setting_paginate_by

from ..classes import Setting, SettingNamespace

from .literals import (
    ENVIRONMENT_TEST_NAME, ENVIRONMENT_TEST_VALUE, TEST_SETTING_GLOBAL_NAME,
    TEST_SETTING_INITIAL_VALUE, TEST_SETTING_VALIDATION_BAD_VALUE,
    TEST_SETTING_VALIDATION_GOOD_VALUE, TEST_SETTING_VALUE
)
from .mixins import SmartSettingNamespaceTestMixin, SmartSettingTestMixin
from .mocks import (
    TestNamespaceMigrationInvalid, TestNamespaceMigrationInvalidDual,
    TestNamespaceMigrationOne, TestNamespaceMigrationTwo,
    test_validation_function
)


class SettingNamespaceMigrationTestCase(
    SmartSettingNamespaceTestMixin, SmartSettingTestMixin, BaseTestCase
):
    def test_environment_migration(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(TEST_SETTING_GLOBAL_NAME),
            value=TEST_SETTING_INITIAL_VALUE
        )
        self._create_test_settings_namespace(
            migration_class=TestNamespaceMigrationOne, version='0002'
        )
        self._create_test_setting()

        self.assertEqual(
            self.test_setting.value, TEST_SETTING_INITIAL_VALUE
        )

    def test_migration_0001_to_0002(self):
        self._create_test_settings_namespace(
            migration_class=TestNamespaceMigrationTwo, version='0002'
        )
        self._create_test_setting()

        self.test_config_value = TEST_SETTING_VALUE
        self._create_test_config_file()

        self.assertEqual(
            self.test_setting.value, '{}_0001'.format(TEST_SETTING_VALUE)
        )

    def test_migration_0001_to_0003(self):
        self._create_test_settings_namespace(
            migration_class=TestNamespaceMigrationTwo, version='0003'
        )
        self._create_test_setting()

        self.test_config_value = TEST_SETTING_VALUE
        self._create_test_config_file()

        self.assertEqual(
            self.test_setting.value, '{}_0001_0002'.format(TEST_SETTING_VALUE)
        )

    def test_migration_invalid(self):
        self._create_test_settings_namespace(
            migration_class=TestNamespaceMigrationInvalid, version='0002'
        )
        self._create_test_setting()

        self.test_config_value = TEST_SETTING_VALUE
        self._create_test_config_file()

        self.assertEqual(
            self.test_setting.value, TEST_SETTING_VALUE
        )

    def test_migration_invalid_dual(self):
        self._create_test_settings_namespace(
            migration_class=TestNamespaceMigrationInvalidDual, version='0002'
        )
        self._create_test_setting()

        self.test_config_value = TEST_SETTING_VALUE
        self._create_test_config_file()

        self.assertEqual(
            self.test_setting.value, TEST_SETTING_VALUE
        )


class SettingTestCase(
    SmartSettingTestMixin, SmartSettingNamespaceTestMixin, BaseTestCase
):
    def test_environment_override(self):
        test_environment_value = 'test environment value'
        test_file_value = 'test file value'

        self._create_test_settings_namespace()
        self._create_test_setting()

        self._set_environment_variable(
            name='MAYAN_{}'.format(self.test_setting.global_name),
            value=test_environment_value
        )

        self.test_config_value = test_file_value
        self._create_test_config_file()

        self.assertEqual(
            self.test_setting.value, test_environment_value
        )

    def test_environment_variable(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(ENVIRONMENT_TEST_NAME),
            value=ENVIRONMENT_TEST_VALUE
        )

        self.assertEqual(
            setting_paginate_by.value, int(ENVIRONMENT_TEST_VALUE)
        )

    def test_config_backup_creation(self):
        path_config_backup = Path(settings.CONFIGURATION_LAST_GOOD_FILEPATH)
        fs_cleanup(
            filename=str(path_config_backup)
        )

        Setting.save_last_known_good()
        self.assertTrue(
            path_config_backup.exists()
        )

    def test_config_backup_creation_no_tags(self):
        path_config_backup = Path(settings.CONFIGURATION_LAST_GOOD_FILEPATH)
        fs_cleanup(
            filename=str(path_config_backup)
        )

        Setting.save_last_known_good()
        self.assertTrue(
            path_config_backup.exists()
        )

        with path_config_backup.open(mode='r') as file_object:
            self.assertFalse(
                '!!python/' in file_object.read()
            )

    def test_setting_check_changed(self):
        self._create_test_settings_namespace()
        test_setting = self.test_settings_namespace.add_setting(
            global_name='SMART_SETTINGS_TEST_SETTING',
            default='test value'
        )
        # Initialize hash cache.
        Setting._cache_hash = None
        Setting.check_changed()
        self.assertFalse(
            Setting.check_changed()
        )
        test_setting.value = 'test value edited'
        self.assertTrue(
            Setting.check_changed()
        )

    def test_setting_check_changed_for_translations(self):
        self._create_test_settings_namespace()
        self.test_settings_namespace.add_setting(
            global_name='SMART_SETTINGS_TEST_SETTING',
            default=_('English')
        )
        # Initialize hash cache.
        Setting._cache_hash = None
        Setting.check_changed()
        self.assertFalse(
            Setting.check_changed()
        )

        translation.activate(language='es')
        self.assertFalse(
            Setting.check_changed()
        )
        translation.activate(language='en')

    def test_setting_validation_call(self):
        self._create_test_settings_namespace()
        self._create_test_setting(
            validation_function=test_validation_function
        )

        self.test_setting.validate(
            raw_value=TEST_SETTING_VALIDATION_GOOD_VALUE
        )

        with self.assertRaises(expected_exception=ValidationError):
            self.test_setting.validate(
                raw_value=TEST_SETTING_VALIDATION_BAD_VALUE
            )

    def test_setting_validation_at_loading(self):
        self._create_test_settings_namespace()
        self._create_test_setting(
            validation_function=test_validation_function
        )

        SettingNamespace.invalidate_cache_all()

        self.test_config_value = TEST_SETTING_VALIDATION_BAD_VALUE
        self._create_test_config_file()

        with self.assertRaises(expected_exception=ValidationError):
            self.test_setting.value



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins import IndexTemplateTestMixin


class IndexTemplateTestCase(IndexTemplateTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_method_get_absolute_url(self):
        self._create_test_index_template()

        self._clear_events()

        self.assertTrue(
            self._test_index_template.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('dynamic_search', '0001_initial')
    ]

    operations = [
        migrations.AlterField(
            model_name='recentsearch',
            name='user',
            field=models.ForeignKey(
                editable=False, on_delete=models.CASCADE,
                to=settings.AUTH_USER_MODEL, verbose_name='User'
            ),
            preserve_default=True
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0064_auto_20201012_0544'),
        ('document_parsing', '0007_auto_20200917_0736'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentPageContent',
            new_name='DocumentFilePageContent',
        ),
    ]



from mayan.apps.source_interactive.source_backends.mixins import (
    SourceBackendMixinInteractive
)
from mayan.apps.sources.source_backends.base import SourceBackend

from ..source_backends.mixins import SourceBackendMixinCompressed

from .source_backend_actions import (
    SourceBackendActionDocumentUploadBasicInteractiveCompressed
)


class SourceBackendTestCompressed(
    SourceBackendMixinCompressed, SourceBackendMixinInteractive, SourceBackend
):
    action_class_list = (
        SourceBackendActionDocumentUploadBasicInteractiveCompressed,
    )
    label = 'Test source backend compressed'

    def get_view_context(self, context, request):
        return {
            'subtemplates_list': [
                {
                    'context': {
                        'forms': context['forms'],
                        'is_multipart': True
                    },
                    'name': 'appearance/generic_multiform_subtemplate.html'
                }
            ]
        }



from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Comment

from .literals import TEST_COMMENT_TEXT, TEST_COMMENT_TEXT_EDITED


class DocumentCommentTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Comment
    _test_object_name = '_test_document_comment'

    def _create_test_comment(self, text=None, user=None):
        text = text or TEST_COMMENT_TEXT
        user = user or self._test_user

        self._test_document_comment = self._test_document.comments.create(
            text=text, user=user
        )


class DocumentCommentAPIViewTestMixin(DocumentCommentTestMixin):
    def _request_test_comment_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:comment-list', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'text': TEST_COMMENT_TEXT
            }
        )

        self._test_object_set()

        return response

    def _request_test_comment_delete_api_view(self):
        return self.delete(
            viewname='rest_api:comment-detail', kwargs={
                'document_id': self._test_document.pk,
                'comment_id': self._test_document_comment.pk
            }
        )

    def _request_test_comment_detail_api_view(self):
        return self.get(
            viewname='rest_api:comment-detail', kwargs={
                'document_id': self._test_document.pk,
                'comment_id': self._test_document_comment.pk
            }
        )

    def _request_test_comment_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:comment-detail', kwargs={
                'document_id': self._test_document.pk,
                'comment_id': self._test_document_comment.pk
            }, data={'text': TEST_COMMENT_TEXT_EDITED}
        )

    def _request_test_comment_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:comment-detail', kwargs={
                'document_id': self._test_document.pk,
                'comment_id': self._test_document_comment.pk
            }, data={'text': TEST_COMMENT_TEXT_EDITED}
        )

    def _request_test_comment_list_api_view(self):
        return self.get(
            viewname='rest_api:comment-list', kwargs={
                'document_id': self._test_document.pk
            }
        )


class DocumentCommentViewTestMixin(DocumentCommentTestMixin):
    def _request_test_comment_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='comments:comment_add', kwargs={
                'document_id': self._test_document.pk
            }, data={'text': TEST_COMMENT_TEXT}
        )

        self._test_object_set()

        return response

    def _request_test_comment_delete_view(self):
        return self.post(
            viewname='comments:comment_delete', kwargs={
                'comment_id': self._test_document_comment.pk
            }
        )

    def _request_test_comment_detail_view(self):
        return self.get(
            viewname='comments:comment_details', kwargs={
                'comment_id': self._test_document_comment.pk
            }
        )

    def _request_test_comment_edit_view(self):
        return self.post(
            viewname='comments:comment_edit', kwargs={
                'comment_id': self._test_document_comment.pk
            }, data={
                'text': TEST_COMMENT_TEXT_EDITED
            }
        )

    def _request_test_comment_list_view(self):
        return self.get(
            viewname='comments:comments_for_document', kwargs={
                'document_id': self._test_document.pk
            }
        )



from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_error_log_deleted
from ..models import ErrorLogPartitionEntry

from .mixins import ErrorLogPartitionEntryTestMixin


class ErrorLoggingModelTestCase(ErrorLogPartitionEntryTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_error_log_object()
        self._create_test_error_log_entry()

    def test_entries_limit(self):
        self.error_log.limit = 3
        self._test_object.error_log.create(text='1')
        self._test_object.error_log.create(text='2')
        self._test_object.error_log.create(text='3')
        self._test_object.error_log.create(text='4')

        self.assertEqual(
            list(
                self._test_object.error_log.values_list('text', flat=True)
            ), ['2', '3', '4']
        )

    def test_orphan_error_logs_after_object_deletion(self):
        self._clear_events()

        self._test_object.delete()
        self.assertEqual(
            ErrorLogPartitionEntry.objects.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_error_log_deletion(self):
        self._clear_events()

        self._test_error_log_entry.delete()
        self.assertEqual(
            ErrorLogPartitionEntry.objects.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_object)
        self.assertEqual(events[0].target, self._test_object)
        self.assertEqual(events[0].verb, event_error_log_deleted.id)



from django.db import migrations

from mayan.apps.documents import settings
from mayan.apps.storage.classes import DefinedStorage

STORAGE_NAME_DOCUMENT_IMAGE = 'documents__documentimagecache'


def code_purge_and_delete_file_cache(apps, schema_editor):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    try:
        cache = Cache.objects.get(
            defined_storage_name=STORAGE_NAME_DOCUMENT_IMAGE
        )
    except Cache.DoesNotExist:
        return
    else:
        try:
            DefinedStorage.get(name=cache.defined_storage_name)
        except KeyError:
            """
            Unknown or deleted storage. Must not be purged otherwise only
            the database data will be erased but the actual storage files
            will remain.
            """
        else:
            for partition in cache.partitions.all():

                for parition_file in partition.files.all():
                    parition_file.delete()

        cache.delete()


def code_purge_and_delete_file_cache_reverse(apps, schema_editor):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    # Protect against future rename of setting_document_cache_maximum_size
    setting_document_cache_maximum_size = getattr(
        settings, 'setting_document_cache_maximum_size', None
    )
    if setting_document_cache_maximum_size:
        document_cache_maximum_size = setting_document_cache_maximum_size.value
    else:
        document_cache_maximum_size = 500 * 2 ** 20  # 500MB default

    Cache.objects.create(
        defined_storage_name=STORAGE_NAME_DOCUMENT_IMAGE,
        maximum_size=document_cache_maximum_size
    )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0055_auto_20200814_0626'),
        ('file_caching', '0006_auto_20200322_0626')
    ]

    operations = [
        migrations.RunPython(
            code=code_purge_and_delete_file_cache,
            reverse_code=code_purge_and_delete_file_cache_reverse
        )
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.model_mixins import BackendModelMixin
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..classes import WorkflowActionNull
from ..events import event_workflow_template_edited
from ..literals import WORKFLOW_ACTION_ON_ENTRY, WORKFLOW_ACTION_WHEN_CHOICES

from .workflow_state_action_model_mixins import WorkflowStateActionBusinessLogicMixin
from .workflow_state_models import WorkflowState

__all__ = ('WorkflowStateAction',)


class WorkflowStateAction(
    WorkflowStateActionBusinessLogicMixin, BackendModelMixin,
    ExtraDataModelMixin, models.Model
):
    _backend_model_null_backend = WorkflowActionNull

    state = models.ForeignKey(
        on_delete=models.CASCADE, related_name='actions', to=WorkflowState,
        verbose_name=_('Workflow state')
    )
    label = models.CharField(
        max_length=255, help_text=_('A short text describing the action.'),
        verbose_name=_('Label')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )
    when = models.PositiveIntegerField(
        choices=WORKFLOW_ACTION_WHEN_CHOICES,
        default=WORKFLOW_ACTION_ON_ENTRY, help_text=_(
            'At which moment of the state this action will execute.'
        ), verbose_name=_('When')
    )
    condition = models.TextField(
        blank=True, help_text=_(
            'The condition that will determine if this state action '
            'is executed or not. The condition is evaluated against the '
            'workflow instance. Conditions that do not return any value, '
            'that return the Python logical None, or an empty string (\'\') '
            'are considered to be logical false, any other value is '
            'considered to be the logical true.'
        ), verbose_name=_('Condition')
    )

    class Meta:
        ordering = ('label',)
        unique_together = ('state', 'label')
        verbose_name = _('Workflow state action')
        verbose_name_plural = _('Workflow state actions')

    def __str__(self):
        return self.label

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='state.workflow',
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'state.workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'state.workflow'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



class WorkflowException(Exception):
    """Base exception for the document states app"""


class WorkflowStateActionError(WorkflowException):
    """Raise for errors during execution of workflow state actions"""



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..events import event_workflow_instance_transitioned

from .mixins.workflow_template_state_escalation_mixins import WorkflowTemplateStateEscalationTaskTestMixin


class WorkflowTemplateStateEscalationTaskTestCase(
    WorkflowTemplateStateEscalationTaskTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_workflow_template_state_escalation(
            extra_kwargs={'unit': 'microseconds'}
        )
        self._create_test_document_stub()

    def test_task_workflow_instance_check_escalation(self):
        test_workflow_instance = self._test_document.workflows.first()
        test_workflow_instance_state = test_workflow_instance.get_current_state()

        self._clear_events()

        self._execute_task_workflow_instance_check_escalation(
            test_workflow_instance_id=test_workflow_instance.pk
        )

        self.assertNotEqual(
            test_workflow_instance.get_current_state(),
            test_workflow_instance_state
        )
        self.assertEqual(
            test_workflow_instance.get_last_log_entry().comment,
            self._test_workflow_template_state_escalation.comment
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].verb, event_workflow_instance_transitioned.id
        )

    def test_task_workflow_instance_check_escalation_all(self):
        test_workflow_instance = self._test_document.workflows.first()
        test_workflow_instance_state = test_workflow_instance.get_current_state()

        self._clear_events()

        self._execute_task_workflow_instance_check_escalation_all()

        self.assertNotEqual(
            test_workflow_instance.get_current_state(),
            test_workflow_instance_state
        )
        self.assertEqual(
            test_workflow_instance.get_last_log_entry().comment,
            self._test_workflow_template_state_escalation.comment
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].verb, event_workflow_instance_transitioned.id
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.rest_api import serializers


class AJAXTemplateSerializer(serializers.Serializer):
    hex_hash = serializers.CharField(
        label=_('Hex hash'), read_only=True
    )
    name = serializers.CharField(
        label=_('Name'), read_only=True
    )
    html = serializers.CharField(
        label=_('HTML'), read_only=True
    )
    url = serializers.HyperlinkedIdentityField(
        label=_('URL'), lookup_field='name', lookup_url_kwarg='name',
        view_name='rest_api:template-detail'
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_search, icon_search_advanced, icon_search_again,
    icon_search_backend_reindex
)
from .permissions import permission_search_tools
from .search_backends import SearchBackend


def condition_search_backend_supports_reindexing(context, resolved_object):
    search_backend_class = SearchBackend.get_class()
    return search_backend_class.feature_reindex


link_search = Link(
    args='search_model.full_name', icon=icon_search,
    text=_('Basic search'), view='search:search_simple'
)
link_search_advanced = Link(
    args='search_model.full_name', icon=icon_search_advanced,
    text=_('Advanced search'), view='search:search_advanced'
)
link_search_again = Link(
    args='search_model.full_name', icon=icon_search_again,
    keep_query=True, text=_('Search again'), view='search:search_again'
)
link_search_backend_reindex = Link(
    condition=condition_search_backend_supports_reindexing,
    icon=icon_search_backend_reindex, permissions=(permission_search_tools,),
    text=_('Reindex search backend'), view='search:search_backend_reindex'
)



from mayan.apps.document_states import storages
from mayan.apps.smart_settings.tests.mixins import SmartSettingTestMixin
from mayan.apps.storage.tests.mixins import StorageSettingTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import STORAGE_NAME_WORKFLOW_CACHE
from ..settings import setting_workflow_image_cache_storage_backend_arguments


class WorkflowStorageSettingsTestCase(
    SmartSettingTestMixin, StorageSettingTestMixin, BaseTestCase
):
    def test_setting_workflow_image_cache_storage_backend_arguments_invalid_value(self):
        assertion = self._test_storage_setting_with_invalid_value(
            setting=setting_workflow_image_cache_storage_backend_arguments,
            storage_module=storages,
            storage_name=STORAGE_NAME_WORKFLOW_CACHE
        )

        self.assertTrue(
            'Unable to initialize' in str(assertion.exception)
        )
        self.assertTrue(
            'workflow preview' in str(assertion.exception)
        )



from django.dispatch import Signal

signal_post_document_type_change = Signal(
    providing_args=('instance',), use_caching=True
)
signal_post_initial_document_type = Signal(
    providing_args=('instance',), use_caching=True
)
signal_post_document_file_upload = Signal(
    providing_args=('instance',), use_caching=True
)
signal_post_document_version_remap = Signal(
    providing_args=('instance',), use_caching=True
)



import json

from django.utils.module_loading import import_string

from mayan.apps.credentials.tests.mixins import StoredCredentialPasswordUsernameTestMixin
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import UserMailer

from .literals import (
    TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS, MAILER_BACKEND_TEST_PATH,
    TEST_USER_MAILER_LABEL
)


class MailerTestMixin(
    StoredCredentialPasswordUsernameTestMixin, TestMixinObjectCreationTrack
):
    _test_mailer_backend_path = MAILER_BACKEND_TEST_PATH
    _test_object_model = UserMailer
    _test_object_name = '_test_user_mailer'

    def setUp(self):
        super().setUp()
        self._test_mailer_backend_class = self._get_test_mailer_backend_class()

    def _get_test_mailer_backend_class(self):
        return import_string(dotted_path=self._test_mailer_backend_path)

    def _create_test_user_mailer(self, extra_backend_data=None):
        backend_data = {'from': TEST_EMAIL_FROM_ADDRESS}

        if extra_backend_data:
            backend_data.update(**extra_backend_data)

        self._test_user_mailer = UserMailer.objects.create(
            default=True,
            enabled=True,
            label=TEST_USER_MAILER_LABEL,
            backend_path=self._test_mailer_backend_class.backend_id,
            backend_data=json.dumps(obj=backend_data)
        )


class DocumentMailerViewTestMixin(MailerTestMixin):
    def _request_test_document_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_link_single', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_link_multiple', query={
                'id_list': self._test_document.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )


class DocumentFileMailerViewTestMixin(MailerTestMixin):
    def _request_test_document_file_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_single', kwargs={
                'document_file_id': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_multiple', query={
                'id_list': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_single', kwargs={
                'document_file_id': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_file_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_multiple', query={
                'id_list': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )


class DocumentVersionMailerViewTestMixin(MailerTestMixin):
    def _request_test_document_version_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_single', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_version_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_multiple', query={
                'id_list': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            },
        )

    def _request_test_document_version_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_single',
            kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )

    def _request_test_document_version_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_multiple',
            query={
                'id_list': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'user_mailer': self._test_user_mailer.pk
            }
        )


class MailerViewTestMixin(MailerTestMixin):
    def _request_test_user_mailer_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='mailer:user_mailer_create', kwargs={
                'backend_path': self._test_mailer_backend_class.backend_id
            }, data={
                'default': True,
                'enabled': True,
                'label': TEST_USER_MAILER_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_user_mailer_delete_view(self):
        return self.post(
            viewname='mailer:user_mailer_delete', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }
        )

    def _request_test_user_mailer_edit_view(self):
        return self.post(
            viewname='mailer:user_mailer_edit', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }, data={
                'label': '{}_edited'.format(TEST_USER_MAILER_LABEL)
            }
        )

    def _request_test_user_mailer_list_view(self):
        return self.get(
            viewname='mailer:user_mailer_list'
        )

    def _request_test_user_mailer_log_entry_view(self):
        return self.get(
            viewname='mailer:user_mailer_log', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }
        )

    def _request_test_user_mailer_test_view(self):
        return self.post(
            viewname='mailer:user_mailer_test', kwargs={
                'mailer_id': self._test_user_mailer.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                )
            }
        )



import logging

from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import SmartLinkConditionForm
from ..icons import (
    icon_smart_link_condition, icon_smart_link_condition_create,
    icon_smart_link_condition_delete, icon_smart_link_condition_edit,
    icon_smart_link_condition_list
)
from ..links import link_smart_link_condition_create
from ..models import SmartLink, SmartLinkCondition
from ..permissions import permission_smart_link_edit

logger = logging.getLogger(name=__name__)


class SmartLinkConditionListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = SmartLink
    external_object_permission = permission_smart_link_edit
    external_object_pk_url_kwarg = 'smart_link_id'
    view_icon = icon_smart_link_condition_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_smart_link_condition,
            'no_results_main_link': link_smart_link_condition_create.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                'Conditions are small logic units that when combined '
                'define how the smart link will behave.'
            ),
            'no_results_title': _(
                'There are no conditions for this smart link'
            ),
            'object': self.external_object,
            'title': _(
                'Conditions for smart link: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.conditions.all()


class SmartLinkConditionCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = SmartLink
    external_object_permission = permission_smart_link_edit
    external_object_pk_url_kwarg = 'smart_link_id'
    form_class = SmartLinkConditionForm
    view_icon = icon_smart_link_condition_create

    def get_extra_context(self):
        return {
            'title': _(
                'Add new conditions to smart link: "%s"'
            ) % self.external_object,
            'object': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'smart_link': self.external_object
        }

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'smart_link_id': self.external_object.pk
            }, viewname='linking:smart_link_condition_list'
        )

    def get_queryset(self):
        return self.external_object.conditions.all()


class SmartLinkConditionDeleteView(SingleObjectDeleteView):
    model = SmartLinkCondition
    object_permission = permission_smart_link_edit
    pk_url_kwarg = 'smart_link_condition_id'
    view_icon = icon_smart_link_condition_delete

    def get_extra_context(self):
        return {
            'condition': self.object,
            'navigation_object_list': ('object', 'condition'),
            'object': self.object.smart_link,
            'title': _(
                'Delete smart link condition: "%s"?'
            ) % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'smart_link_id': self.object.smart_link.pk
            }, viewname='linking:smart_link_condition_list'
        )


class SmartLinkConditionEditView(SingleObjectEditView):
    form_class = SmartLinkConditionForm
    model = SmartLinkCondition
    object_permission = permission_smart_link_edit
    pk_url_kwarg = 'smart_link_condition_id'
    view_icon = icon_smart_link_condition_edit

    def get_extra_context(self):
        return {
            'condition': self.object,
            'navigation_object_list': ('object', 'condition'),
            'object': self.object.smart_link,
            'title': _('Edit smart link condition')
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'smart_link_id': self.object.smart_link.pk
            }, viewname='linking:smart_link_condition_list'
        )



import logging

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.db import models

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.serialization import yaml_load

from .classes import Layer
from .transformations import BaseTransformation

logger = logging.getLogger(name=__name__)


class LayerTransformationManager(models.Manager):
    def get_for_object(
        self, obj, as_classes=False, maximum_layer_order=None,
        only_stored_layer=None, user=None
    ):
        """
        as_classes == True returns the transformation classes from
        `.classes` ready to be feed to the converter class.
        """
        Layer.update()

        StoredLayer = apps.get_model(
            app_label='converter', model_name='StoredLayer'
        )
        content_type = ContentType.objects.get_for_model(model=obj)

        transformations = self.filter(
            enabled=True, object_layer__content_type=content_type,
            object_layer__enabled=True, object_layer__object_id=obj.pk
        )

        if maximum_layer_order is not None:
            access_layers = StoredLayer.objects.filter(
                order__lte=maximum_layer_order
            )
            exclude_layers = StoredLayer.objects.filter(
                order__gt=maximum_layer_order
            )
        else:
            access_layers = StoredLayer.objects.all()
            exclude_layers = StoredLayer.objects.none()

        for stored_layer in access_layers:
            try:
                layer_class = stored_layer.get_layer()
            except KeyError:
                """
                This was a class defined but later erased. Ignore it.
                """
            else:
                access_permission = layer_class.permissions.get(
                    'access', None
                )
                if access_permission:
                    try:
                        AccessControlList.objects.check_access(
                            obj=obj, permissions=(access_permission,),
                            user=user
                        )
                    except PermissionDenied:
                        access_layers = access_layers.exclude(
                            pk=stored_layer.pk
                        )

        for stored_layer in exclude_layers:
            exclude_permission = stored_layer.get_layer().permissions.get(
                'exclude', None
            )
            if exclude_permission:
                try:
                    AccessControlList.objects.check_access(
                        obj=obj, permissions=(exclude_permission,), user=user
                    )
                except PermissionDenied:
                    exclude_layers = exclude_layers.exclude(
                        pk=stored_layer.pk
                    )
                    access_layers |= StoredLayer.objects.filter(
                        pk=stored_layer.pk
                    )

        if only_stored_layer:
            transformations = transformations.filter(
                object_layer__stored_layer=only_stored_layer
            )

        transformations = transformations.filter(
            object_layer__stored_layer__in=access_layers
        )

        transformations = transformations.exclude(
            object_layer__stored_layer__in=exclude_layers
        )

        if as_classes:
            result = []
            for transformation in transformations:
                try:
                    transformation_class = BaseTransformation.get(
                        transformation.name
                    )
                except KeyError:
                    # Non existent transformation, but we don't raise an
                    # error.
                    logger.error(
                        'Non existent transformation: %s for %s',
                        transformation.name, obj
                    )
                else:
                    try:
                        # Some transformations don't require arguments
                        # return an empty dictionary as ** doesn't allow None
                        if transformation.arguments:
                            kwargs = yaml_load(
                                stream=transformation.arguments
                            )
                        else:
                            kwargs = {}

                        tranformation_instance = transformation_class(
                            **kwargs
                        )
                        tranformation_instance.object_layer = transformation.object_layer
                        result.append(
                            tranformation_instance
                        )
                    except Exception as exception:
                        logger.error(
                            'Error while parsing transformation "%s", '
                            'arguments "%s", for object "%s"; %s',
                            transformation, transformation.arguments, obj,
                            exception, exc_info=True
                        )

            return result
        else:
            return transformations


class ObjectLayerManager(models.Manager):
    def get_for(self, layer, obj):
        content_type = ContentType.objects.get_for_model(model=obj)

        return self.get_or_create(
            content_type=content_type, object_id=obj.pk,
            stored_layer=layer.stored_layer
        )



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


def code_add_user_options_to_existing_users(apps, schema_editor):
    User = apps.get_model(settings.AUTH_USER_MODEL)
    UserOptions = apps.get_model(
        app_label='user_management', model_name='UserOptions'
    )

    for user in User.objects.using(alias=schema_editor.connection.alias).all():
        UserOptions.objects.using(
            alias=schema_editor.connection.alias
        ).create(user=user)


def code_remove_user_options_from_existing_users(apps, schema_editor):
    User = apps.get_model(settings.AUTH_USER_MODEL)
    UserOptions = apps.get_model(
        app_label='user_management', model_name='UserOptions'
    )

    for user in User.objects.using(alias=schema_editor.connection.alias).all():
        UserOptions.objects.using(
            alias=schema_editor.connection.alias
        ).filter(user=user).delete()


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='UserOptions',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'block_password_change', models.BooleanField(
                        default=False,
                        verbose_name='Forbid this user from changing their '
                        'password.'
                    )
                ),
                (
                    'user', models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='user_options',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                )
            ],
            options={
                'verbose_name': 'User settings',
                'verbose_name_plural': 'Users settings'
            },
        ),
        migrations.RunPython(
            code=code_add_user_options_to_existing_users,
            reverse_code=code_remove_user_options_from_existing_users
        )
    ]



from django.contrib.auth.models import Group
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import event_role_created, event_role_edited
from .managers import RoleManager, StoredPermissionManager
from .model_mixins import (
    RoleBusinessLogicMixin, StoredPermissionBusinessLogicMixin
)


class Role(ExtraDataModelMixin, RoleBusinessLogicMixin, models.Model):
    """
    This model represents a Role. Roles are permission units. They are the
    only object to which permissions can be granted. They are themselves
    containers too, containing Groups, which are organization units. Roles
    are the basic method to grant a permission to a group. Permissions granted
    to a group using a role, are granted for the entire system.
    """
    label = models.CharField(
        help_text=_('A short text describing the role.'), max_length=128,
        unique=True, verbose_name=_('Label')
    )
    permissions = models.ManyToManyField(
        related_name='roles', to='StoredPermission',
        verbose_name=_('Permissions')
    )
    groups = models.ManyToManyField(
        related_name='roles', to=Group, verbose_name=_('Groups')
    )

    objects = RoleManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Role')
        verbose_name_plural = _('Roles')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='permissions:role_list')

    def natural_key(self):
        return (self.label,)
    natural_key.dependencies = ['auth.Group', 'permissions.StoredPermission']

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_role_created,
            'target': 'self',
        },
        edited={
            'event': event_role_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class StoredPermission(StoredPermissionBusinessLogicMixin, models.Model):
    """
    This model is the counterpart of the permissions.classes.Permission
    class. Allows storing a database counterpart of a permission class.
    It is used to store the permissions help by a role or in an ACL.
    """
    namespace = models.CharField(
        max_length=64, verbose_name=_('Namespace')
    )
    name = models.CharField(
        max_length=64, verbose_name=_('Name')
    )

    objects = StoredPermissionManager()

    class Meta:
        ordering = ('namespace',)
        unique_together = ('namespace', 'name')
        verbose_name = _('Permission')
        verbose_name_plural = _('Permissions')

    def __str__(self):
        return str(self.label)

    def get_absolute_url(self):
        return reverse(
            viewname='permissions:stored_permission_detail', kwargs={
                'stored_permission_id': self.pk
            }
        )

    def natural_key(self):
        return (self.namespace, self.name)



from django.test import override_settings

from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_file_metadata_document_file_finished,
    event_file_metadata_document_file_submitted
)
from ..permissions import (
    permission_document_type_file_metadata_setup,
    permission_file_metadata_submit, permission_file_metadata_view
)

from .literals import TEST_FILE_METADATA_KEY
from .mixins import DocumentTypeViewTestMixin, FileMetadataViewTestMixin


class DocumentTypeViewTestCase(
    DocumentTypeViewTestMixin, GenericDocumentViewTestCase
):
    def test_document_type_settings_view_no_permission(self):
        self._clear_events()

        response = self._request_document_type_file_metadata_settings_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_settings_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_file_metadata_setup
        )

        self._clear_events()

        response = self._request_document_type_file_metadata_settings_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_submit_view_no_permission(self):
        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_type_file_metadata_submit_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_submit_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type, permission=permission_file_metadata_submit
        )

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_type_file_metadata_submit_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(
            events[0].verb, event_file_metadata_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(
            events[1].verb, event_file_metadata_document_file_finished.id
        )

    def test_trashed_document_document_type_submit_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_file_metadata_submit
        )

        self._test_document.delete()

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_type_file_metadata_submit_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


@override_settings(FILE_METADATA_AUTO_PROCESS=True)
class FileMetadataViewTestCase(
    FileMetadataViewTestMixin, GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        self.test_driver = self._test_document.file_latest.file_metadata_drivers.first()

    def test_document_file_driver_list_view_no_permission(self):
        self._clear_events()

        response = self._request_document_file_metadata_driver_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_driver_list_view_with_access(self):
        self.grant_access(
            permission=permission_file_metadata_view, obj=self._test_document
        )

        self._clear_events()

        response = self._request_document_file_metadata_driver_list_view()
        self.assertContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_driver_list_view_with_access(self):
        self.grant_access(
            permission=permission_file_metadata_view, obj=self._test_document
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_file_metadata_driver_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_file_metadata_list_view_no_permission(self):
        self._clear_events()

        response = self._request_document_file_metadata_list_view()
        self.assertNotContains(
            response=response, text=TEST_FILE_METADATA_KEY, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_file_metadata_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_file_metadata_view
        )

        self._clear_events()

        response = self._request_document_file_metadata_list_view()
        self.assertContains(
            response=response, text=TEST_FILE_METADATA_KEY, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_metadata_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_file_metadata_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_file_metadata_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_submit_view_no_permission(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_file_metadata_single_submit_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_submit_single_view_with_access(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()
        self.grant_access(
            permission=permission_file_metadata_submit, obj=self._test_document
        )

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_file_metadata_single_submit_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(
            events[0].verb, event_file_metadata_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(
            events[1].verb, event_file_metadata_document_file_finished.id
        )

    def test_trashed_document_file_submit_view_with_access(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()

        self.grant_access(
            permission=permission_file_metadata_submit, obj=self._test_document
        )

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_file_metadata_single_submit_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_submit_multiple_view_no_permission(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_file_multiple_submit_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_submit_multiple_view_with_access(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()
        self.grant_access(
            permission=permission_file_metadata_submit, obj=self._test_document
        )

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._clear_events()

        response = self._request_document_file_multiple_submit_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(
            events[0].verb, event_file_metadata_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(
            events[1].verb, event_file_metadata_document_file_finished.id
        )

    def test_trashed_document_file_submit_multiple_view_with_access(self):
        self._test_document.file_latest.file_metadata_drivers.all().delete()

        self.grant_access(
            permission=permission_file_metadata_submit, obj=self._test_document
        )

        file_metadata_driver_count = self._test_document.file_latest.file_metadata_drivers.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_file_multiple_submit_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.file_metadata_drivers.count(),
            file_metadata_driver_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.http import Http404

from mayan.apps.views.generics import (
    SingleObjectDynamicFormCreateView, SingleObjectDynamicFormEditView
)

from .view_mixins import ViewMixinDynamicFormBackendClass


class ViewSingleObjectDynamicFormModelBackendCreate(
    ViewMixinDynamicFormBackendClass, SingleObjectDynamicFormCreateView
):
    def get_backend_class(self):
        try:
            return self.backend_class.get(
                name=self.kwargs['backend_path']
            )
        except KeyError:
            raise Http404(
                '{} class not found'.format(
                    self.kwargs['backend_path']
                )
            )


class ViewSingleObjectDynamicFormModelBackendEdit(
    ViewMixinDynamicFormBackendClass, SingleObjectDynamicFormEditView
):
    def get_backend_class(self):
        return self.object.get_backend_class()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .classes import DocumentFileCompressor
from .literals import (
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY,
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT
)

namespace = SettingNamespace(
    label=_('Document downloads'), name='document_downloads'
)

setting_message_body_template = namespace.add_setting(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY,
    global_name='DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY', help_text=_(
        'Template for the document download message body text. Can '
        'include HTML. Available variables: {}'.format(
            ', '.join(DocumentFileCompressor.context_key_list)
        )
    )
)
setting_message_subject_template = namespace.add_setting(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT,
    global_name='DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT', help_text=_(
        'Template for the document download message subject line. '
        'Can\'t include HTML. Available variables: {}'.format(
            ', '.join(DocumentFileCompressor.context_key_list)
        )
    )
)



from contextlib import contextmanager
import logging

from django.apps import apps
from django.core.files.base import ContentFile
from django.db.models import F, Sum
from django.template.defaultfilters import filesizeformat
from django.utils.functional import cached_property
from django.utils.text import format_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.decorators import (
    acquire_lock_class_method, locked_class_method, release_lock_class_method
)
from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.storage.classes import DefinedStorage

from .events import event_cache_partition_purged, event_cache_purged
from .exceptions import FileCachingException
from .settings import (
    setting_maximum_failed_prune_attempts,
    setting_maximum_normal_prune_attempts
)

logger = logging.getLogger(name=__name__)


class CacheBusinessLogicMixin:
    def get_files(self):
        CachePartitionFile = apps.get_model(
            app_label='file_caching', model_name='CachePartitionFile'
        )

        return CachePartitionFile.objects.filter(
            partition__cache__id=self.pk
        )

    def get_maximum_size_display(self):
        return filesizeformat(bytes_=self.maximum_size)

    get_maximum_size_display.help_text = _(
        'Size at which the cache will start deleting old entries.'
    )
    get_maximum_size_display.short_description = _('Maximum size')

    def get_defined_storage(self):
        try:
            defined_storage_class = DefinedStorage.get(
                name=self.defined_storage_name
            )
        except KeyError:
            defined_storage_class = DefinedStorage(
                dotted_path='', label=_('Unknown'), name='unknown'
            )

        return defined_storage_class

    def get_partition_count(self):
        CachePartition = apps.get_model(
            app_label='file_caching', model_name='CachePartition'
        )

        queryset_cache_partitions = CachePartition.objects.filter(cache=self)

        return queryset_cache_partitions.count()

    get_partition_count.short_description = _('Partition count')
    get_partition_count.help_text = _('Total cached objects.')

    def get_partition_file_count(self):
        queryset_files = self.get_files()
        return queryset_files.count()

    get_partition_file_count.short_description = _('Partition file count')
    get_partition_file_count.help_text = _('Total cached files.')

    def get_total_size(self):
        """
        Return the actual usage of the cache.
        """
        queryset_files = self.get_files()
        queryset_files_aggregated = queryset_files.aggregate(
            file_size__sum=Sum('file_size')
        )

        return queryset_files_aggregated['file_size__sum'] or 0

    def get_total_size_display(self):
        total_size = self.get_total_size()

        size_humanized = filesizeformat(
            bytes_=total_size
        )

        size_percent = total_size / self.maximum_size * 100

        return format_lazy(
            '{} ({:0.1f}%)', size_humanized, size_percent
        )

    get_total_size_display.short_description = _('Current size')
    get_total_size_display.help_text = _('Current size of the cache.')

    @cached_property
    def label(self):
        defined_storage_class = self.get_defined_storage()
        return defined_storage_class.label

    def prune(self):
        """
        Deletes files until the total size of the cache is below the allowed
        maximum size of the cache.
        """
        failed_attempts = 0
        normal_attempts = 0
        file_index = 0

        CachePartitionFile = apps.get_model(
            app_label='file_caching', model_name='CachePartitionFile'
        )

        while self.get_total_size() >= self.maximum_size:
            queryset_cache_partition_file = self.get_files().order_by(
                'hits', 'datetime'
            )

            try:
                cache_partition_file = queryset_cache_partition_file[
                    file_index
                ]
            except IndexError:
                # Attempted to get a file beyond what the queryset provided.
                file_index = 0
            else:
                try:
                    cache_partition_file.delete()
                except CachePartitionFile.DoesNotExist:
                    # The file selected from deletion was deleted by another
                    # process before the lock was acquired.
                    file_index += 1
                except LockError:
                    logger.debug(
                        'Lock error trying to delete file "%s" for '
                        'prune. Skipping and attempting next file.',
                        cache_partition_file
                    )
                    failed_attempts += 1
                    file_index += 1

                    if failed_attempts > setting_maximum_failed_prune_attempts.value:
                        raise FileCachingException(
                            'Too many cache prune attempts failed.'
                        )
                else:
                    file_index = 0
                    normal_attempts += 1

                    if normal_attempts > setting_maximum_normal_prune_attempts.value:
                        raise FileCachingException(
                            'Too many cache prunes trying to create a '
                            'single new file.'
                        )

    @method_event(
        event=event_cache_purged,
        event_manager_class=EventManagerMethodAfter,
        target='self'
    )
    def purge(self, user):
        """
        Deletes the entire cache.
        """
        self._event_actor = user

        try:
            DefinedStorage.get(name=self.defined_storage_name)
        except KeyError:
            """
            Unknown or deleted storage. Must not be purged otherwise only
            the database data will be erased but the actual storage files
            will remain.
            """
        else:
            queryset_partitions = self.partitions.all()

            for partition in queryset_partitions:
                partition._event_action_object = self
                try:
                    partition.purge(user=user)
                except Exception as exception:
                    logger.error(
                        'Unable to purge partition ID: %d; %s',
                        partition.pk, exception
                    )
                    # Don't raise exceptions to allow the loop to continue and
                    # avoid a single exception from stopping the purge.

    @cached_property
    def storage(self):
        defined_storage_class = self.get_defined_storage()
        defined_storage_instance = defined_storage_class.get_storage_instance()

        return defined_storage_instance


class CachePartitionBusinessLogicMixin:
    @staticmethod
    def get_combined_filename(parent, filename):
        return '{}-{}'.format(parent, filename)

    def _lock_manager_get_lock_name(self, filename):
        return self.get_file_lock_name(filename=filename)

    @contextmanager
    def create_file(self, filename):
        lock_name = self.get_file_lock_name(filename=filename)
        try:
            logger.debug('trying to acquire lock: %s', lock_name)
            locking_backend_class = LockingBackend.get_backend()
            lock = locking_backend_class.acquire_lock(name=lock_name)
            logger.debug('acquired lock: %s', lock_name)
            try:
                self.cache.prune()

                # Since open "wb+" doesn't create files, force the creation
                # of an empty file.
                try:
                    self.cache.storage.delete(
                        name=self.get_full_filename(filename=filename)
                    )
                except Exception as exception:
                    """
                    Some S3 implementations like Google Cloud Storage throw
                    an error when attempting to delete a not existent file
                    key. Ignore this exception, any storage error of concern
                    will be triggered by the ``storage.save`` call below.
                    """
                    logger.debug(
                        'cache.storage.delete exception: %s', exception
                    )

                self.cache.storage.save(
                    name=self.get_full_filename(filename=filename),
                    content=ContentFile(content=b'')
                )

                partition_file = None

                try:
                    partition_file = self.files.create(filename=filename)
                    yield partition_file._open_for_writing(_acquire_lock=False)
                except Exception as exception:
                    logger.error(
                        'Unexpected exception while trying to save new '
                        'cache file; %s', exception, exc_info=True
                    )
                    if partition_file:
                        partition_file.delete(_acquire_lock=False)
                    else:
                        # If the CachePartitionFile entry was not created
                        # do manual clean up of the empty storage file
                        # created with the previous`self.cache.storage.save`.
                        self.cache.storage.delete(
                            name=self.get_full_filename(filename=filename)
                        )
                    raise
                else:
                    partition_file.close(_acquire_lock=False)
                    partition_file._update_size(_acquire_lock=False)
            finally:
                lock.release()
        except LockError:
            logger.debug('unable to obtain lock: %s', lock_name)
            raise

    def get_file(self, filename):
        return self.files.get(filename=filename)

    def get_file_lock_name(self, filename):
        return 'cache_partition-file-{}-{}-{}'.format(
            self.cache.pk, self.pk, filename
        )

    def get_full_filename(self, filename):
        CachePartition = apps.get_model(
            app_label='file_caching', model_name='CachePartition'
        )

        return CachePartition.get_combined_filename(
            parent=self.name, filename=filename
        )

    def get_total_size(self):
        """
        Return the actual usage of the cache partition.
        """
        queryset_files = self.files
        queryset_files_aggregated = queryset_files.aggregate(
            file_size__sum=Sum('file_size')
        )

        return queryset_files_aggregated['file_size__sum'] or 0

    def get_total_size_display(self):
        return filesizeformat(
            bytes_=self.get_total_size()
        )

    get_total_size_display.short_description = _('Current size')
    get_total_size_display.help_text = _(
        'Current size of the cache partition.'
    )

    @method_event(
        event=event_cache_partition_purged,
        event_manager_class=EventManagerMethodAfter,
        target='self'
    )
    def purge(self, user):
        self._event_actor = user
        queryset_files = self.files.all()

        for parition_file in queryset_files:
            try:
                parition_file.delete()
            except Exception as exception:
                logger.error(
                    'Unable to delete partition file ID: %d; %s',
                    parition_file.pk, exception
                )
                # Don't raise exceptions to allow the loop to continue and
                # avoid a single exception from stopping the purge.


class CachePartitionFileBusinessLogicMixin:
    def _lock_manager_get_lock_name(self, *args, **kwargs):
        return self.partition.get_file_lock_name(filename=self.filename)

    @acquire_lock_class_method
    def _open_for_writing(self):
        """
        Open the file for writing only. If the file is written to, the
        .update_size() must be called.
        """
        try:
            storage_instance = self.partition.cache.storage

            self._storage_object = storage_instance.open(
                mode='wb', name=self.full_filename
            )
            return self._storage_object
        except Exception as exception:
            logger.error(
                'Unexpected exception opening the cache file; %s', exception,
                exc_info=True
            )
            raise

    @locked_class_method
    def _update_size(self):
        """
        Called after creation and initial write only.
        """
        storage_instance = self.partition.cache.storage
        self.file_size = storage_instance.size(name=self.full_filename)
        self.save(
            update_fields=('file_size',)
        )
        if self.file_size > self.partition.cache.maximum_size:
            raise FileCachingException(
                'Cache partition file %s is bigger than the maximum cache '
                'size.'
            )

    @release_lock_class_method
    def close(self):
        if self._storage_object is not None:
            self._storage_object.close()
        self._storage_object = None

    @cached_property
    def full_filename(self):
        CachePartition = apps.get_model(
            app_label='file_caching', model_name='CachePartition'
        )

        return CachePartition.get_combined_filename(
            parent=self.partition.name, filename=self.filename
        )

    @contextmanager
    def open(self):
        """
        Open the file for reading only.
        """
        CachePartitionFile = apps.get_model(
            app_label='file_caching', model_name='CachePartitionFile'
        )

        lock_name = self._lock_manager_get_lock_name()
        try:
            logger.debug('trying to acquire lock: %s', lock_name)
            locking_backend_class = LockingBackend.get_backend()

            self._lock = locking_backend_class.acquire_lock(name=lock_name)

            queryset_partition_files = CachePartitionFile.objects.filter(
                pk=self.pk
            )

            queryset_partition_files.update(
                hits=F('hits') + 1
            )
            logger.debug('acquired lock: %s', lock_name)
            self._storage_object = None
            try:
                storage_instance = self.partition.cache.storage
                self._storage_object = storage_instance.open(
                    mode='rb', name=self.full_filename
                )
            except Exception as exception:
                logger.error(
                    'Unexpected exception opening the cache file; %s',
                    exception, exc_info=True
                )
                raise
            else:
                yield self._storage_object
            finally:
                self.close(_acquire_lock=False)
                self._lock.release()
        except LockError:
            logger.debug('unable to obtain lock: %s', lock_name)
            raise



from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_metadata_type_created, event_metadata_type_edited,
    event_metadata_type_relationship_updated
)
from ..models import MetadataType
from ..permissions import (
    permission_metadata_type_create, permission_metadata_type_delete,
    permission_metadata_type_edit, permission_metadata_type_view
)

from .mixins.metadata_type_mixins import MetadataTypeViewTestMixin


class DocumentTypeMetadataTypeRelationshipViewTestCase(
    MetadataTypeViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type()

    def test_document_type_relationship_delete_view_no_permission(self):
        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type
        )

        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self._clear_events()

        response = self._request_test_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_delete_view_with_metadata_type_access(self):
        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type
        )

        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_delete_view_with_document_type_access(self):
        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type
        )

        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_delete_view_with_full_access(self):
        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type
        )

        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(
            events[0].verb, event_metadata_type_relationship_updated.id
        )

    def test_document_type_relationship_edit_view_no_permission(self):
        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self._clear_events()

        response = self._request_test_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_edit_view_with_metadata_type_access(self):
        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_edit_view_with_document_type_access(self):
        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_relationship_edit_view_with_full_access(self):
        test_document_type_metadata_type_relationship_count = self._test_document_type.metadata.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.metadata.count(),
            test_document_type_metadata_type_relationship_count + 1
        )

        self.assertQuerysetEqual(
            qs=self._test_document_type.metadata.values(
                'metadata_type', 'required'
            ),
            values=[
                {
                    'metadata_type': self._test_metadata_type.pk,
                    'required': True,
                }
            ], transform=dict
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(
            events[0].verb, event_metadata_type_relationship_updated.id
        )


class MetadataTypeViewTestCase(
    MetadataTypeViewTestMixin, GenericViewTestCase
):
    def test_metadata_type_create_view_no_permission(self):
        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_create_view_with_access(self):
        self.grant_permission(permission=permission_metadata_type_create)
        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_metadata_type)
        self.assertEqual(events[0].verb, event_metadata_type_created.id)

    def test_metadata_type_single_delete_view_no_permission(self):
        self._create_test_metadata_type()
        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_single_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_single_delete_view_with_access(self):
        self._create_test_metadata_type()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_delete
        )
        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_single_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_multiple_delete_view_no_permission(self):
        self._create_test_metadata_type()

        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_multiple_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_multiple_delete_view_with_access(self):
        self._create_test_metadata_type()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_delete
        )

        metadata_type_count = MetadataType.objects.count()

        self._clear_events()

        response = self._request_test_metadata_type_multiple_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            MetadataType.objects.count(), metadata_type_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_edit_view_no_permission(self):
        self._create_test_metadata_type()
        metadata_type_values = self._model_instance_to_dictionary(
            instance=self._test_metadata_type
        )

        self._clear_events()

        response = self._request_test_metadata_type_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_metadata_type.refresh_from_db()
        self.assertEqual(
            self._model_instance_to_dictionary(
                instance=self._test_metadata_type
            ), metadata_type_values
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_edit_view_with_access(self):
        self._create_test_metadata_type()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )
        metadata_type_values = self._model_instance_to_dictionary(
            instance=self._test_metadata_type
        )

        self._clear_events()

        response = self._request_test_metadata_type_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_metadata_type.refresh_from_db()
        self.assertNotEqual(
            self._model_instance_to_dictionary(
                instance=self._test_metadata_type
            ), metadata_type_values
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_metadata_type)
        self.assertEqual(events[0].verb, event_metadata_type_edited.id)

    def test_metadata_type_list_view_no_permission(self):
        self._create_test_metadata_type()

        self._clear_events()

        response = self._request_metadata_type_list_view()
        self.assertNotContains(
            response=response, status_code=200, text=self._test_metadata_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_list_view_with_access(self):
        self._create_test_metadata_type()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_view
        )

        self._clear_events()

        response = self._request_metadata_type_list_view()
        self.assertContains(
            response=response, status_code=200, text=self._test_metadata_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class MetadataTypeDocumentTypeRelationshipViewTestCase(
    MetadataTypeViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type()

    def test_metadata_type_relationship_delete_view_no_permission(self):
        self._test_metadata_type.document_types.create(
            document_type=self._test_document_type
        )

        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 404)

        self._test_metadata_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_relationship_delete_view_with_document_type_access(self):
        self._test_metadata_type.document_types.create(
            document_type=self._test_document_type
        )

        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 404)

        self._test_metadata_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_relationship_delete_view_with_metadata_type_access(self):
        self._test_metadata_type.document_types.create(
            document_type=self._test_document_type
        )

        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 302)

        self._test_metadata_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_relationship_delete_view_with_full_access(self):
        self._test_metadata_type.document_types.create(
            document_type=self._test_document_type
        )

        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_delete_view()
        self.assertEqual(response.status_code, 302)

        self._test_metadata_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(
            events[0].verb, event_metadata_type_relationship_updated.id
        )

    def test_metadata_type_document_type_relationship_view_no_permission(self):
        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_document_type_relationship_view_with_document_type_access(self):
        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_document_type_relationship_view_with_metadata_type_access(self):
        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_document_type_relationship_view_with_full_access(self):
        test_metadata_type_document_type_relationship_count = self._test_metadata_type.document_types.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_metadata_type_edit
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_metadata_type_document_type_relationship_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_metadata_type.document_types.count(),
            test_metadata_type_document_type_relationship_count + 1
        )

        self.assertQuerysetEqual(
            qs=self._test_document_type.metadata.values(
                'metadata_type', 'required'
            ),
            values=[
                {
                    'metadata_type': self._test_metadata_type.pk,
                    'required': True,
                }
            ], transform=dict
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(
            events[0].verb, event_metadata_type_relationship_updated.id
        )



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_multi_item, menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.logging.classes import ErrorLog
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.column_widgets import TwoStateWidget

from .classes import MailerBackend
from .events import event_email_sent, event_profile_edited
from .links import (
    link_send_document_file_attachment_multiple,
    link_send_document_file_attachment_single,
    link_send_document_file_link_multiple,
    link_send_document_file_link_single, link_send_document_link_multiple,
    link_send_document_link_single,
    link_send_document_version_attachment_multiple,
    link_send_document_version_attachment_single,
    link_send_document_version_link_multiple,
    link_send_document_version_link_single, link_user_mailer_create,
    link_user_mailer_delete, link_user_mailer_edit, link_user_mailer_list,
    link_user_mailer_setup, link_user_mailer_test
)
from .permissions import (
    permission_send_document_file_attachment,
    permission_send_document_file_link, permission_send_document_link,
    permission_send_document_version_attachment,
    permission_send_document_version_link, permission_user_mailer_delete,
    permission_user_mailer_edit, permission_user_mailer_use,
    permission_user_mailer_view
)


class MailerApp(MayanAppConfig):
    app_namespace = 'mailer'
    app_url = 'mailer'
    has_tests = True
    name = 'mayan.apps.mailer'
    verbose_name = _('Mailer')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        UserMailer = self.get_model(model_name='UserMailer')

        error_log = ErrorLog(app_config=self)
        error_log.register_model(model=UserMailer, register_permission=True)

        EventModelRegistry.register(model=UserMailer)

        MailerBackend.load_modules()

        ModelCopy(
            model=UserMailer, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'label', 'enabled', 'backend_path', 'backend_data'
            ), field_values={
                'default': False
            }
        )

        ModelEventType.register(
            model=UserMailer, event_types=(
                event_email_sent, event_profile_edited
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=UserMailer
        )
        SourceColumn(
            attribute='default', include_label=True, is_sortable=True,
            source=UserMailer, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=UserMailer, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_backend_class_label', include_label=True,
            source=UserMailer
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_send_document_link,
            )
        )
        ModelPermission.register(
            model=DocumentFile, permissions=(
                permission_send_document_file_attachment,
                permission_send_document_file_link
            )
        )
        ModelPermission.register(
            model=DocumentVersion, permissions=(
                permission_send_document_version_attachment,
                permission_send_document_version_link
            )
        )

        ModelPermission.register(
            model=UserMailer, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_user_mailer_delete, permission_user_mailer_edit,
                permission_user_mailer_view, permission_user_mailer_use
            )
        )

        # Document

        menu_multi_item.bind_links(
            links=(
                link_send_document_link_multiple,
            ), sources=(Document,)
        )

        menu_object.bind_links(
            links=(
                link_send_document_link_single,
            ), sources=(Document,)
        )

        # Document file

        menu_multi_item.bind_links(
            links=(
                link_send_document_file_attachment_multiple,
                link_send_document_file_link_multiple
            ), sources=(DocumentFile,)
        )

        menu_object.bind_links(
            links=(
                link_send_document_file_link_single,
                link_send_document_file_attachment_single
            ), sources=(DocumentFile,)
        )

        # Document version

        menu_multi_item.bind_links(
            links=(
                link_send_document_version_attachment_multiple,
                link_send_document_version_link_multiple
            ), sources=(DocumentVersion,)
        )

        menu_object.bind_links(
            links=(
                link_send_document_version_link_single,
                link_send_document_version_attachment_single
            ), sources=(DocumentVersion,)
        )

        # Mailing profile

        menu_object.bind_links(
            links=(
                link_user_mailer_delete, link_user_mailer_edit,
                link_user_mailer_test
            ), sources=(UserMailer,)
        )

        menu_return.bind_links(
            links=(link_user_mailer_list,), sources=(
                UserMailer, 'mailer:user_mailer_list',
                'mailer:user_mailer_create'
            )
        )

        menu_secondary.bind_links(
            links=(link_user_mailer_create,), sources=(
                UserMailer, 'mailer:user_mailer_list',
                'mailer:user_mailer_create'
            )
        )

        menu_setup.bind_links(
            links=(link_user_mailer_setup,)
        )



from django.db import migrations


def code_document_file_download_event_and_permission_update(apps, schema_editor):
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )
    StoredPermission = apps.get_model(
        app_label='permissions', model_name='StoredPermission'
    )
    Role = apps.get_model(
        app_label='permissions', model_name='Role'
    )

    Action.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        verb='documents.document_file_downloaded'
    ).update(
        verb='document_downloads.document_file_downloaded'
    )

    StoredEventType.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        name='documents.document_file_downloaded'
    ).delete()

    try:
        stored_permission = StoredPermission.objects.using(
            alias=schema_editor.connection.alias
        ).get(
            namespace='document_downloads', name='document_file_download'
        )
    except StoredPermission.DoesNotExist:
        """Raised on initial migrations. Can be ignored."""
    else:
        AccessControlList.permissions.through.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            storedpermission__namespace='documents',
            storedpermission__name='document_file_download'
        ).update(storedpermission_id=stored_permission.pk)

        Role.permissions.through.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            storedpermission__namespace='documents',
            storedpermission__name='document_file_download'
        ).update(
            storedpermission_id=stored_permission.pk
        )

    StoredPermission.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        namespace='documents', name='document_file_download'
    ).delete()


def code_document_file_download_event_and_permission_update_reverse(apps, schema_editor):
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )
    StoredPermission = apps.get_model(
        app_label='permissions', model_name='StoredPermission'
    )
    Role = apps.get_model(
        app_label='permissions', model_name='Role'
    )

    Action.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        verb='document_downloads.document_file_downloaded'
    ).update(
        verb='documents.document_file_downloaded'
    )

    StoredEventType.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        name='document_downloads.document_file_downloaded'
    ).delete()

    stored_permission = StoredPermission.objects.using(
        alias=schema_editor.connection.alias
    ).create(
        namespace='documents', name='document_file_download'
    )

    AccessControlList.permissions.through.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        storedpermission__namespace='document_downloads',
        storedpermission__name='document_file_download'
    ).update(storedpermission_id=stored_permission.pk)

    Role.permissions.through.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        storedpermission__namespace='document_downloads',
        storedpermission__name='document_file_download'
    ).update(
        storedpermission_id=stored_permission.pk
    )

    StoredPermission.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        namespace='document_downloads', name='document_file_download'
    ).delete()


class Migration(migrations.Migration):
    dependencies = [
        ('acls', '0004_auto_20210130_0322'),
        ('documents', '0077_favoritedocumentproxy'),
        ('events', '0008_auto_20180315_0029'),
        ('permissions', '0004_auto_20191213_0044')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_file_download_event_and_permission_update,
            reverse_code=code_document_file_download_event_and_permission_update_reverse
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('acls', '0002_auto_20150703_0513')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='accesscontrollist',
            options={
                'ordering': ('pk',), 'verbose_name': 'Access entry',
                'verbose_name_plural': 'Access entries'
            }
        )
    ]



from django.http import Http404

from actstream.models import Action, any_stream

from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalContentTypeObjectAPIViewMixin

from .classes import EventType, EventTypeNamespace
from .models import Notification
from .permissions import permission_events_view
from .serializers import (
    EventSerializer, EventTypeNamespaceSerializer, EventTypeSerializer,
    NotificationSerializer
)


class APIObjectEventListView(
    ExternalContentTypeObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of events for the specified object.
    """
    mayan_external_object_permissions = {
        'GET': (permission_events_view,)
    }
    ordering_fields = ('id', 'timestamp')
    serializer_class = EventSerializer

    def get_source_queryset(self):
        return any_stream(
            obj=self.get_external_object()
        )


class APIEventTypeNamespaceDetailView(generics.RetrieveAPIView):
    """
    get: Returns the details of an event type namespace.
    """
    serializer_class = EventTypeNamespaceSerializer

    def get_object(self):
        try:
            return EventTypeNamespace.get(
                name=self.kwargs['name']
            )
        except KeyError:
            raise Http404


class APIEventTypeNamespaceListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event type namespaces.
    """
    serializer_class = EventTypeNamespaceSerializer
    source_queryset = EventTypeNamespace.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }


class APIEventTypeNamespaceEventTypeListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event types from a namespaces.
    """
    serializer_class = EventTypeSerializer

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }

    def get_source_queryset(self):
        try:
            return EventTypeNamespace.get(
                name=self.kwargs['name']
            ).get_event_types()
        except KeyError:
            raise Http404


class APIEventTypeListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event types.
    """
    serializer_class = EventTypeSerializer
    source_queryset = EventType.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }


class APIEventListView(generics.ListAPIView):
    """
    get: Returns a list of all the available events.
    """
    mayan_view_permissions = {
        'GET': (permission_events_view,)
    }
    ordering_fields = ('id', 'timestamp')
    serializer_class = EventSerializer
    source_queryset = Action.objects.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }


class APINotificationListView(generics.ListAPIView):
    """
    get: Return a list of notifications for the current user.
    """
    ordering_fields = ('action__timestamp', 'id', 'read')
    serializer_class = NotificationSerializer

    def get_source_queryset(self):
        parameter_read = self.request.GET.get('read')

        if self.request.user.is_authenticated:
            queryset = Notification.objects.filter(user=self.request.user)
        else:
            queryset = Notification.objects.none()

        if parameter_read == 'True':
            queryset = queryset.filter(read=True)
        elif parameter_read == 'False':
            queryset = queryset.filter(read=False)

        return queryset



from django.shortcuts import get_object_or_404

from rest_framework.generics import get_object_or_404 as rest_get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_file_new
from mayan.apps.documents.tasks import task_document_upload

from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)

from .arguments import (
    argument_document, argument_document_id, argument_document_id_optional
)
from .document_description_mixins import SourceBackendActionMixinDocumentDescriptionInteractive
from .document_label_mixins import SourceBackendActionMixinLabelInteractive
from .document_language_mixins import SourceBackendActionMixinLanguageInteractive
from .user_mixins import SourceBackendActionMixinUserInteractive


class SourceBackendActionMixinDocumentInteractive:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                document = argument_document

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['document'] = self.context['document']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                document_id = argument_document_id

            def process_interface_context(self):
                super().process_interface_context()

                queryset = AccessControlList.objects.restrict_queryset(
                    queryset=Document.valid.all(),
                    permission=permission_document_file_new,
                    user=self.context['request'].user
                )

                self.action_kwargs['document'] = rest_get_object_or_404(
                    queryset=queryset, pk=self.context['document_id']
                )

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                document_id = argument_document_id

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['document'] = Document.valid.get(
                    pk=self.context['document_id']
                )

        class View(SourceBackendActionInterfaceRequestViewForm):
            class Argument:
                document = argument_document

            def process_interface_context(self):
                super().process_interface_context()

                queryset = AccessControlList.objects.restrict_queryset(
                    queryset=Document.valid.all(),
                    permission=permission_document_file_new,
                    user=self.context['request'].user
                )

                self.action_kwargs['document'] = get_object_or_404(
                    klass=queryset, pk=self.context['document'].pk
                )

    def _background_task(self, document, **kwargs):
        result = super()._background_task(**kwargs)

        result['document'] = document

        return result

    def get_task_kwargs(self, document, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['document_id'] = document.pk

        return result


class SourceBackendActionMixinDocumentOptionalTaskOnly:
    class Interface:
        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                document_id = argument_document_id_optional

            def process_interface_context(self):
                super().process_interface_context()

                document_id = self.context['document_id']

                if document_id:
                    self.action_kwargs['document'] = Document.valid.get(
                        pk=document_id
                    )

    def _background_task(self, document=None, **kwargs):
        result = super()._background_task(**kwargs)

        result['document'] = document

        return result

    def get_document_file_task_kwargs(
        self, server_upload_entry, document=None, **kwargs
    ):
        result = super().get_document_file_task_kwargs(
            server_upload_entry=server_upload_entry, **kwargs
        )

        if document:
            result['document_id'] = document.pk

        return result

    def get_task_kwargs(self, document=None, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        if document:
            result['action_interface_kwargs']['document_id'] = document.pk

        return result


class SourceBackendActionMixinDocumentUploadBase:
    def _background_task(self, **kwargs):
        result = super()._background_task(**kwargs)

        if result:
            # Make this optional in case another mixin interrupted the MRO,
            # And called a background task.
            server_upload_entry_list = result.get(
                'server_upload_entry_list', ()
            )

            for server_upload_entry in server_upload_entry_list:
                document_task_kwargs = self.get_document_task_kwargs(
                    server_upload_entry=server_upload_entry, **kwargs
                )

                document_task_kwargs['shared_uploaded_file_id'] = server_upload_entry[
                    'shared_uploaded_file_id'
                ]

                task_document_upload.apply_async(
                    kwargs=document_task_kwargs
                )

    def get_document_task_kwargs(
        self, server_upload_entry, document_type, **kwargs
    ):
        result = super().get_document_task_kwargs(
            server_upload_entry=server_upload_entry,
            **kwargs
        )

        result['document_type_id'] = document_type.pk

        return result


class SourceBackendActionMixinDocumentUploadInteractive(
    SourceBackendActionMixinDocumentDescriptionInteractive,
    SourceBackendActionMixinLabelInteractive,
    SourceBackendActionMixinLanguageInteractive,
    SourceBackendActionMixinUserInteractive,
    SourceBackendActionMixinDocumentUploadBase
):
    """
    Mixin for a complete action that uploads documents.
    """


class SourceBackendActionMixinDocumentUploadInteractiveNot(
    SourceBackendActionMixinDocumentUploadBase
):
    """
    Same as `SourceBackendActionMixinDocumentUploadBase` but subclassed for
    clarity.
    """



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Cache',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'name', models.CharField(
                        max_length=128, unique=True, verbose_name='Name'
                    )
                ),
                (
                    'label', models.CharField(
                        max_length=128, verbose_name='Label'
                    )
                ),
                (
                    'maximum_size', models.PositiveIntegerField(
                        verbose_name='Maximum size'
                    )
                ),
                (
                    'storage_instance_path', models.CharField(
                        max_length=255, unique=True,
                        verbose_name='Storage instance path'
                    )
                )
            ],
            options={
                'verbose_name': 'Cache',
                'verbose_name_plural': 'Caches'
            }
        ),
        migrations.CreateModel(
            name='CachePartition',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'name', models.CharField(
                        max_length=128, verbose_name='Name'
                    )
                ),
                (
                    'cache', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='partitions', to='file_caching.Cache',
                        verbose_name='Cache'
                    )
                )
            ],
            options={
                'verbose_name': 'Cache partition',
                'verbose_name_plural': 'Cache partitions'
            }
        ),
        migrations.CreateModel(
            name='CachePartitionFile',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, db_index=True,
                        verbose_name='Date time'
                    )
                ),
                (
                    'filename', models.CharField(
                        max_length=255, verbose_name='Filename'
                    )
                ),
                (
                    'file_size', models.PositiveIntegerField(
                        default=0, verbose_name='File size'
                    )
                ),
                (
                    'partition', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='files',
                        to='file_caching.CachePartition',
                        verbose_name='Cache partition'
                    )
                )
            ],
            options={
                'get_latest_by': 'datetime',
                'verbose_name': 'Cache partition file',
                'verbose_name_plural': 'Cache partition files'
            },
        ),
        migrations.AlterUniqueTogether(
            name='cachepartitionfile',
            unique_together={
                ('partition', 'filename')
            }
        ),
        migrations.AlterUniqueTogether(
            name='cachepartition',
            unique_together={
                ('cache', 'name')
            }
        )
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.classes import DefinedStorage

from .literals import (
    STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE, STORAGE_NAME_DOCUMENT_FILES,
    STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
)
from .settings import (
    setting_document_file_page_image_cache_storage_backend,
    setting_document_file_page_image_cache_storage_backend_arguments,
    setting_document_file_storage_backend,
    setting_document_file_storage_backend_arguments,
    setting_document_version_page_image_cache_storage_backend,
    setting_document_version_page_image_cache_storage_backend_arguments
)

storage_document_files = DefinedStorage(
    dotted_path=setting_document_file_storage_backend.value,
    error_message=_(
        'Unable to initialize the document file storage. Check '
        'the settings {} and {} for formatting errors.'.format(
            setting_document_file_storage_backend.global_name,
            setting_document_file_storage_backend_arguments.global_name
        )
    ),
    label=_('Document files'),
    name=STORAGE_NAME_DOCUMENT_FILES,
    kwargs=setting_document_file_storage_backend_arguments.value
)

storage_document_file_image_cache = DefinedStorage(
    dotted_path=setting_document_file_page_image_cache_storage_backend.value,
    error_message=_(
        'Unable to initialize the document file image storage. Check '
        'the settings {} and {} for formatting errors.'.format(
            setting_document_file_page_image_cache_storage_backend.global_name,
            setting_document_file_page_image_cache_storage_backend_arguments.global_name
        )
    ),
    label=_('Document file page images'),
    name=STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE,
    kwargs=setting_document_file_page_image_cache_storage_backend_arguments.value
)

storage_document_version_image_cache = DefinedStorage(
    dotted_path=setting_document_version_page_image_cache_storage_backend.value,
    error_message=_(
        'Unable to initialize the document version image storage. Check '
        'the settings {} and {} for formatting errors.'.format(
            setting_document_version_page_image_cache_storage_backend.global_name,
            setting_document_version_page_image_cache_storage_backend_arguments.global_name
        )
    ),
    label=_('Document version page images'),
    name=STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE,
    kwargs=setting_document_version_page_image_cache_storage_backend_arguments.value
)



from django.shortcuts import get_object_or_404

from rest_framework.generics import (
    get_object_or_404 as rest_get_object_or_404
)

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_create

from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)

from .arguments import argument_document_type, argument_document_type_id


class SourceBackendActionMixinDocumentTypeBase:
    def _background_task(self, document_type, **kwargs):
        result = super()._background_task(**kwargs)

        result['document_type'] = document_type

        return result


class SourceBackendActionMixinDocumentTypeInteractive(
    SourceBackendActionMixinDocumentTypeBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                document_type = argument_document_type

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['document_type'] = self.context['document_type']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                document_type_id = argument_document_type_id

            def process_interface_context(self):
                super().process_interface_context()

                queryset = AccessControlList.objects.restrict_queryset(
                    queryset=DocumentType.objects.all(),
                    permission=permission_document_create,
                    user=self.context['request'].user
                )

                self.action_kwargs['document_type'] = rest_get_object_or_404(
                    queryset=queryset, pk=self.context['document_type_id']
                )

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                document_type_id = argument_document_type_id

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['document_type'] = DocumentType.objects.get(
                    pk=self.context['document_type_id']
                )

        class View(SourceBackendActionInterfaceRequestViewForm):
            class Argument:
                document_type = argument_document_type

            def process_interface_context(self):
                super().process_interface_context()

                queryset = AccessControlList.objects.restrict_queryset(
                    queryset=DocumentType.objects.all(),
                    permission=permission_document_create,
                    user=self.context['request'].user
                )

                self.action_kwargs['document_type'] = get_object_or_404(
                    klass=queryset, pk=self.context['document_type'].pk
                )

    def get_task_kwargs(self, document_type, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['document_type_id'] = document_type.pk

        return result


class SourceBackendActionMixinDocumentTypeInteractiveNot(
    SourceBackendActionMixinDocumentTypeBase
):
    class Interface:
        class Task(SourceBackendActionInterfaceTask):
            def process_interface_context(self):
                super().process_interface_context()

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['document_type'] = source_backend_instance.get_document_type()



from django.contrib.contenttypes.models import ContentType
from django.core.files import File

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..classes import Layer
from ..models import Asset, LayerTransformation
from ..tasks import task_content_object_image_generate
from ..transformations import BaseTransformation

from .literals import (
    TEST_ASSET_INTERNAL_NAME, TEST_ASSET_LABEL, TEST_ASSET_LABEL_EDITED,
    TEST_ASSET_PATH, TEST_LAYER_LABEL, TEST_LAYER_NAME, TEST_LAYER_ORDER,
    TEST_TRANSFORMATION_ARGUMENT, TEST_TRANSFORMATION_ARGUMENT_EDITED,
    TEST_TRANSFORMATION_LABEL, TEST_TRANSFORMATION_NAME
)


class AssetTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Asset
    _test_object_name = '_test_asset'

    def _create_test_asset(self):
        with open(file=TEST_ASSET_PATH, mode='rb') as file_object:
            self._test_asset = Asset.objects.create(
                label=TEST_ASSET_LABEL,
                internal_name=TEST_ASSET_INTERNAL_NAME,
                file=File(file=file_object)
            )


class AssetAPIViewTestMixin(AssetTestMixin):
    def _request_test_asset_create_api_view(self):
        self._test_object_track()

        with open(file=TEST_ASSET_PATH, mode='rb') as file_object:
            response = self.post(
                viewname='rest_api:asset-list', data={
                    'label': TEST_ASSET_LABEL,
                    'internal_name': TEST_ASSET_INTERNAL_NAME,
                    'file': File(file=file_object)
                }
            )

        self._test_object_set()

        return response

    def _request_test_asset_delete_api_view(self):
        return self.delete(
            viewname='rest_api:asset-detail',
            kwargs={'asset_id': self._test_asset.pk}
        )

    def _request_test_asset_detail_api_view(self):
        return self.get(
            viewname='rest_api:asset-detail',
            kwargs={'asset_id': self._test_asset.pk}
        )

    def _request_test_asset_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:asset-detail', kwargs={
                'asset_id': self._test_asset.pk
            }, data={'label': TEST_ASSET_LABEL_EDITED}
        )

    def _request_test_asset_edit_via_put_api_view(self):
        with open(file=TEST_ASSET_PATH, mode='rb') as file_object:
            return self.put(
                viewname='rest_api:asset-detail', kwargs={
                    'asset_id': self._test_asset.pk
                }, data={
                    'label': TEST_ASSET_LABEL_EDITED,
                    'internal_name': TEST_ASSET_INTERNAL_NAME,
                    'file': File(file=file_object)
                }
            )

    def _request_test_asset_list_api_view(self):
        return self.get(viewname='rest_api:asset-list')


class AssetTaskTestMixin(AssetTestMixin):
    def _execute_asset_task_content_object_image_generate(self):
        content_type = ContentType.objects.get_for_model(model=Asset)

        task_content_object_image_generate.apply_async(
            kwargs={
                'content_type_id': content_type.pk,
                'object_id': self._test_asset.pk
            }
        ).get()


class AssetViewTestMixin(AssetTestMixin):
    def _request_test_asset_create_view(self):
        self._test_object_track()

        with open(file=TEST_ASSET_PATH, mode='rb') as file_object:
            response = self.post(
                viewname='converter:asset_create', data={
                    'label': TEST_ASSET_LABEL,
                    'internal_name': TEST_ASSET_INTERNAL_NAME,
                    'file': file_object
                }
            )

        self._test_object_set()

        return response

    def _request_test_asset_delete_view(self):
        return self.post(
            viewname='converter:asset_single_delete', kwargs={
                'asset_id': self._test_asset.pk
            }
        )

    def _request_test_asset_detail_view(self):
        return self.get(
            viewname='converter:asset_detail', kwargs={
                'asset_id': self._test_asset.pk
            }
        )

    def _request_test_asset_edit_view(self):
        return self.post(
            viewname='converter:asset_edit', kwargs={
                'asset_id': self._test_asset.pk
            }, data={
                'label': TEST_ASSET_LABEL_EDITED,
                'internal_name': TEST_ASSET_INTERNAL_NAME
            }
        )

    def _request_test_asset_list_view(self):
        return self.get(viewname='converter:asset_list')


class LayerTestCaseMixin:
    def setUp(self):
        super().setUp()
        Layer.invalidate_cache()


class LayerTestMixin:
    _test_transformation_object = None
    _test_transformation_object_parent = None
    auto_create_test_layer = True

    def setUp(self):
        super().setUp()

        self._test_transformation_object = self._test_document_version_page
        self._test_transformation_object_parent = self._test_document

        self._test_transformation_object_content_type = ContentType.objects.get_for_model(
            model=self._test_transformation_object
        )

        if self.auto_create_test_layer:
            self._create_test_permission()
            self._test_layer_permission_create = self._test_permission

            self._create_test_permission()
            self._test_layer_permission_delete = self._test_permission

            self._create_test_permission()
            self._test_layer_permission_edit = self._test_permission

            self._create_test_permission()
            self._test_layer_permission_view = self._test_permission

            self._test_layer = Layer(
                label=TEST_LAYER_LABEL, name=TEST_LAYER_NAME,
                order=TEST_LAYER_ORDER, permissions={
                    'create': self._test_layer_permission_create,
                    'delete': self._test_layer_permission_delete,
                    'edit': self._test_layer_permission_edit,
                    'select': self._test_layer_permission_create,
                    'view': self._test_layer_permission_view
                }
            )

            ModelPermission.register(
                model=self._test_transformation_object_parent._meta.model,
                permissions=(
                    self._test_layer_permission_create,
                    self._test_layer_permission_delete,
                    self._test_layer_permission_edit,
                    self._test_layer_permission_view
                )
            )

    def tearDown(self):
        if self.auto_create_test_layer:
            Layer._registry.pop(self._test_layer.name, None)

        super().tearDown()


class TransformationTestMixin(LayerTestMixin, TestMixinObjectCreationTrack):
    _test_object_model = LayerTransformation
    _test_object_name = '_test_transformation'
    auto_create_test_transformation_class = True

    def setUp(self):
        super().setUp()
        if self.auto_create_test_transformation_class:
            self._create_test_transformation_class()

        BaseTransformation.register(
            layer=self._test_layer,
            transformation=self.TestTransformationClass
        )

    def _create_test_transformation(self):
        self._test_transformation = self._test_layer.add_transformation_to(
            obj=self._test_transformation_object,
            transformation_class=self.TestTransformationClass,
            arguments=getattr(
                self, '_test_transformation_arguments',
                TEST_TRANSFORMATION_ARGUMENT
            )
        )

    def _create_test_transformation_class(self):
        class TestTransformation(BaseTransformation):
            arguments = ('test_argument',)
            label = TEST_TRANSFORMATION_LABEL
            name = TEST_TRANSFORMATION_NAME

            def execute_on(self, *args, **kwargs):
                super().execute_on(*args, **kwargs)

                return self.image

        self.TestTransformationClass = TestTransformation


class TransformationViewTestMixin(TransformationTestMixin):
    def _request_transformation_create_post_view(self):
        self._test_object_track()

        response = self.post(
            viewname='converter:transformation_create', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name,
                'transformation_name': self.TestTransformationClass.name
            }, data={
                'arguments': getattr(
                    self, '.test_transformation_argument',
                    TEST_TRANSFORMATION_ARGUMENT
                )
            }
        )

        self._test_object_set()

        return response

    def _request_transformation_create_get_view(self):
        return self.get(
            viewname='converter:transformation_create', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name,
                'transformation_name': self.TestTransformationClass.name
            }, data={
                'arguments': getattr(
                    self, '.test_transformation_argument',
                    TEST_TRANSFORMATION_ARGUMENT
                )
            }
        )

    def _request_transformation_delete_view(self):
        return self.post(
            viewname='converter:transformation_delete', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name,
                'transformation_id': self._test_transformation.pk
            }
        )

    def _request_transformation_edit_view(self):
        return self.post(
            viewname='converter:transformation_edit', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name,
                'transformation_id': self._test_transformation.pk
            }, data={
                'arguments': getattr(
                    self, '_test_transformation_argument_edited',
                    TEST_TRANSFORMATION_ARGUMENT_EDITED
                )
            }
        )

    def _request_transformation_list_view(self):
        return self.get(
            viewname='converter:transformation_list', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name
            }
        )

    def _request_test_transformation_select_get_view(self):
        return self.get(
            viewname='converter:transformation_select', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name
            }
        )

    def _request_test_transformation_select_post_view(self):
        return self.post(
            viewname='converter:transformation_select', kwargs={
                'app_label': self._test_transformation_object_content_type.app_label,
                'model_name': self._test_transformation_object_content_type.model,
                'object_id': self._test_transformation_object.pk,
                'layer_name': self._test_layer.name
            }, data={
                'transformation': self.TestTransformationClass.name
            }
        )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_trashed, event_trashed_document_deleted,
    event_trashed_document_restored
)
from ..models.document_models import Document
from ..permissions import (
    permission_document_trash, permission_document_version_view,
    permission_document_view, permission_trashed_document_delete,
    permission_trashed_document_restore
)

from .mixins.document_mixins import DocumentTestMixin
from .mixins.trashed_document_mixins import TrashedDocumentAPIViewTestMixin


class TrashedDocumentAPIViewTestCase(
    TrashedDocumentAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_trash_api_view_no_permission(self):
        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_document_trash_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_trash_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_document_trash_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Document.trash.count(), trashed_document_count + 1)
        self.assertEqual(Document.valid.count(), document_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_trashed.id)

    def test_trashed_document_trash_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        self._test_document.delete()

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_document_trash_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_delete_api_view_no_permission(self):
        self._test_document.delete()

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_delete_api_view_with_access(self):
        self._test_document.delete()
        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_delete
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(Document.trash.count(), trashed_document_count - 1)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_trashed_document_deleted.id)

    def test_non_trashed_document_delete_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_delete
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_detail_api_view_no_permission(self):
        self._test_document.delete()

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertFalse('uuid' in response.data)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_detail_api_view_with_access(self):
        self._test_document.delete()
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['uuid'], str(self._test_document.uuid)
        )

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_non_trashed_document_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_list_api_view_no_permission(self):
        self._test_document.delete()

        self._clear_events()

        response = self._request_test_trashed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_list_api_view_with_access(self):
        self._test_document.delete()
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_trashed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['uuid'],
            str(self._test_document.uuid)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_non_trashed_document_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_trashed_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_via_get_api_view_no_permission(self):
        self._test_document.delete()

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_get_api_view()
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_via_get_api_view_with_access(self):
        self._test_document.delete()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_get_api_view()
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_non_trashed_document_restore_via_get_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_get_api_view()
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_via_post_api_view_no_permission(self):
        self._test_document.delete()

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_post_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_via_post_api_view_with_access(self):
        self._test_document.delete()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_post_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(Document.trash.count(), trashed_document_count - 1)
        self.assertEqual(Document.valid.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_trashed_document_restored.id)

    def test_non_trashed_document_restore_via_post_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        trashed_document_count = Document.trash.count()
        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_via_post_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.trash.count(), trashed_document_count)
        self.assertEqual(Document.valid.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class TrashedDocumentImageAPIViewTestCase(
    TrashedDocumentAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    def test_trashed_document_image_api_view_no_permission(self):
        self._test_document.delete()

        self._clear_events()

        response = self._request_test_trashed_document_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_image_api_view_with_access(self):
        self._test_document.delete()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_trashed_document_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_non_trashed_document_image_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_trashed_document_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.model_mixins import BackendModelMixin
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import event_source_created, event_source_edited
from .managers import SourceManager
from .model_mixins import SourceBusinessLogicMixin
from .source_backends.base import SourceBackendNull


class DocumentFileSourceMetadata(models.Model):
    source = models.ForeignKey(
        on_delete=models.CASCADE, related_name='metadata', to='Source',
        verbose_name=_('Source')
    )
    document_file = models.ForeignKey(
        on_delete=models.CASCADE, related_name='source_metadata',
        to=DocumentFile, verbose_name=_('Document file')
    )
    key = models.CharField(
        db_index=True, help_text=_('Name of the source metadata entry.'),
        max_length=255, verbose_name=_('Key')
    )
    value = models.TextField(
        blank=True, help_text=_(
            'The actual value stored in the source metadata for '
            'the document file.'
        ), null=True, verbose_name=_('Value')
    )

    class Meta:
        ordering = ('key',)
        unique_together = ('source', 'document_file', 'key')
        verbose_name = _('Document file source metadata')
        verbose_name_plural = _('Document file source metadata')


class Source(
    BackendModelMixin, ExtraDataModelMixin, SourceBusinessLogicMixin,
    models.Model
):
    _backend_model_null_backend = SourceBackendNull

    label = models.CharField(
        db_index=True, help_text=_('A short text to describe this source.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )

    objects = SourceManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Source')
        verbose_name_plural = _('Sources')

    def __str__(self):
        return '%s' % self.label

    def delete(self, *args, **kwargs):
        backend_instance = self.get_backend_instance()

        backend_instance.delete()
        super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_source_created,
            'target': 'self'
        },
        edited={
            'event': event_source_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        is_new = not self.pk

        super().save(*args, **kwargs)

        backend_instance = self.get_backend_instance()

        backend_instance.clean()

        if is_new:
            backend_instance.create()
        else:
            backend_instance.update()



from mayan.apps.appearance.classes import Icon

# App

icon_file_metadata = Icon(
    driver_name='fontawesome', symbol='chess-board'
)

# Document type

icon_document_type_file_metadata_settings = Icon(
    driver_name='fontawesome', symbol='chess-board'
)
icon_document_type_file_metadata_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='chess-board',
    secondary_symbol='arrow-right'
)

# Document file

icon_document_file_metadata_single_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='chess-board',
    secondary_symbol='arrow-right'
)
icon_document_file_metadata_multiple_submit = icon_document_file_metadata_single_submit
icon_file_metadata_driver_attribute_list = Icon(
    driver_name='fontawesome', symbol='chess-board'
)
icon_file_metadata_driver_list = Icon(
    driver_name='fontawesome', symbol='chess-board'
)



from django.test import override_settings

from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import (
    TEST_FILE_HYBRID_PDF_CONTENT, TEST_FILE_HYBRID_PDF_FILENAME
)


class DocumentAutoParsingTestCase(GenericDocumentTestCase):
    _test_document_filename = TEST_FILE_HYBRID_PDF_FILENAME
    auto_create_test_document_type = False

    def test_disable_auto_parsing(self):
        self._create_test_document_type()
        self._upload_test_document()
        with self.assertRaises(expected_exception=StopIteration):
            next(
                self._test_document_file.content()
            )

    @override_settings(DOCUMENT_PARSING_AUTO_PARSING=True)
    def test_enabled_auto_parsing(self):
        self._create_test_document_type()
        self._upload_test_document()
        self.assertTrue(
            TEST_FILE_HYBRID_PDF_CONTENT in next(
                self._test_document_file.content()
            )
        )



from ..classes import (
    OriginalDocumentFilenameGenerator, UUIDDocumentFilenameGenerator,
    UUIDPlusOriginalFilename
)

from .base import GenericDocumentTestCase


class DocumentTypeModelFilenameGeneratorTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_original_filename_generator(self):
        self._test_document_type.filename_generator_backend = OriginalDocumentFilenameGenerator.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertEqual(
            str(self._test_document.file_latest.file),
            self._test_document_filename
        )

    def test_uuid_filename_gnerator(self):
        self._test_document_type.filename_generator_backend = UUIDDocumentFilenameGenerator.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertNotEqual(
            str(self._test_document.file_latest.file),
            self._test_document_filename
        )

    def test_uuid_plus_filename_gnerator(self):
        self._test_document_type.filename_generator_backend = UUIDPlusOriginalFilename.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertTrue(
            str(self._test_document.file_latest.file).endswith(
                self._test_document.file_latest.filename
            )
        )
        self.assertFalse(
            str(self._test_document.file_latest.file).startswith(
                self._test_document.file_latest.filename
            )
        )


class DocumentTypeModelTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_method_get_absolute_url(self):
        self.assertTrue(
            self._test_document_type.get_absolute_url()
        )
